@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2001-2009 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -19,12 +19,7 @@
  *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *    Author: Neil Brown
- *    Email: <neilb@cse.unsw.edu.au>
- *    Paper: Neil Brown
- *           School of Computer Science and Engineering
- *           The University of New South Wales
- *           Sydney, 2052
- *           Australia
+ *    Email: <neilb@suse.de>
  *
  *    Additions for bitmap and write-behind RAID options, Copyright (C) 2003-2004,
  *    Paul Clements, SteelEye Technology, Inc.
@@ -46,10 +41,11 @@ int main(int argc, char *argv[])
 
 	int chunk = 0;
 	long long size = -1;
+	long long array_size = -1;
 	int level = UnSet;
 	int layout = UnSet;
+	char *layout_str = NULL;
 	int raiddisks = 0;
-	int max_disks = MD_SB_DISKS; /* just a default */
 	int sparedisks = 0;
 	struct mddev_ident_s ident;
 	char *configfile = NULL;
@@ -91,8 +87,10 @@ int main(int argc, char *argv[])
 
 	char *homehost = NULL;
 	char sys_hostname[256];
+	int require_homehost = 1;
 	char *mailaddr = NULL;
 	char *program = NULL;
+	int increments = 20;
 	int delay = 0;
 	int daemonise = 0;
 	char *pidfile = NULL;
@@ -104,8 +102,8 @@ int main(int argc, char *argv[])
 	int dosyslog = 0;
 	int rebuild_map = 0;
 	int auto_update_home = 0;
+	char *subarray = NULL;
 
-	int copies;
 	int print_help = 0;
 	FILE *outf;
 
@@ -124,13 +122,15 @@ int main(int argc, char *argv[])
 	ident.bitmap_fd = -1;
 	ident.bitmap_file = NULL;
 	ident.name[0] = 0;
+	ident.container = NULL;
+	ident.member = NULL;
 
 	while ((option_index = -1) ,
 	       (opt=getopt_long(argc, argv,
 				shortopt, long_options,
 				&option_index)) != -1) {
 		int newmode = mode;
-		/* firstly, some mode-independant options */
+		/* firstly, some mode-independent options */
 		switch(opt) {
 		case 'h':
 			if (option_index > 0 &&
@@ -151,20 +151,20 @@ int main(int argc, char *argv[])
 			continue;
 
 		case 'b':
-			if (mode == ASSEMBLE || mode == BUILD || mode == CREATE || mode == GROW)
+			if (mode == ASSEMBLE || mode == BUILD || mode == CREATE || mode == GROW ||
+			    mode == INCREMENTAL || mode == MANAGE)
 				break; /* b means bitmap */
 			brief = 1;
-			if (optarg) {
-				fprintf(stderr, Name ": -b cannot have any extra immediately after it, sorry.\n");
-				exit(2);
-			}
 			continue;
 
 		case 'Y': export++;
 			continue;
 
 		case HomeHost:
-			homehost = optarg;
+			if (strcasecmp(optarg, "<ignore>") == 0)
+				require_homehost = 0;
+			else
+				homehost = optarg;
 			continue;
 
 		case ':':
@@ -214,6 +214,17 @@ int main(int argc, char *argv[])
 		case 'o':
 		case 'w':
 		case 'W':
+		case Waitclean:
+		case DetailPlatform:
+		case KillSubarray:
+		case UpdateSubarray:
+			if (opt == KillSubarray || opt == UpdateSubarray) {
+				if (subarray) {
+					fprintf(stderr, Name ": subarray can only be specified once\n");
+					exit(2);
+				}
+				subarray = optarg;
+			}
 		case 'K': if (!mode) newmode = MISC; break;
 		}
 		if (mode && newmode == mode) {
@@ -253,6 +264,7 @@ int main(int argc, char *argv[])
 					dv->writemostly = writemostly;
 					dv->re_add = re_add;
 					dv->used = 0;
+					dv->content = NULL;
 					dv->next = NULL;
 					*devlistend = dv;
 					devlistend = &dv->next;
@@ -261,7 +273,8 @@ int main(int argc, char *argv[])
 					continue;
 				}
 				/* No mode yet, and this is the second device ... */
-				fprintf(stderr, Name ": An option must be given to set the mode before a second device is listed\n");
+				fprintf(stderr, Name ": An option must be given to set the mode before a second device\n"
+					"       (%s) is listed\n", optarg);
 				exit(2);
 			}
 			if (option_index >= 0)
@@ -305,6 +318,8 @@ int main(int argc, char *argv[])
 			dv->disposition = devmode;
 			dv->writemostly = writemostly;
 			dv->re_add = re_add;
+			dv->used = 0;
+			dv->content = NULL;
 			dv->next = NULL;
 			*devlistend = dv;
 			devlistend = &dv->next;
@@ -317,6 +332,7 @@ int main(int argc, char *argv[])
 		 * could depend on the mode */
 #define O(a,b) ((a<<8)|b)
 		switch (O(mode,opt)) {
+		case O(GROW,'c'):
 		case O(CREATE,'c'):
 		case O(BUILD,'c'): /* chunk or rounding */
 			if (chunk) {
@@ -332,9 +348,11 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+#if 0
 		case O(ASSEMBLE,AutoHomeHost):
 			auto_update_home = 1;
 			continue;
+#endif
 		case O(INCREMENTAL, 'e'):
 		case O(CREATE,'e'):
 		case O(ASSEMBLE,'e'):
@@ -350,7 +368,6 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": unrecognised metadata identifier: %s\n", optarg);
 				exit(2);
 			}
-			max_disks = ss->max_devs;
 			continue;
 
 		case O(MANAGE,'W'):
@@ -360,8 +377,15 @@ int main(int argc, char *argv[])
 			writemostly = 1;
 			continue;
 
+		case O(MANAGE,'w'):
+			/* clear write-mostly for following devices */
+			writemostly = 2;
+			continue;
+
+
 		case O(GROW,'z'):
-		case O(CREATE,'z'): /* size */
+		case O(CREATE,'z'):
+		case O(BUILD,'z'): /* size */
 			if (size >= 0) {
 				fprintf(stderr, Name ": size may only be specified once. "
 					"Second value is %s.\n", optarg);
@@ -370,16 +394,36 @@ int main(int argc, char *argv[])
 			if (strcmp(optarg, "max")==0)
 				size = 0;
 			else {
-				size = strtoll(optarg, &c, 10);
-				if (!optarg[0] || *c || size < 4) {
+				size = parse_size(optarg);
+				if (size < 8) {
 					fprintf(stderr, Name ": invalid size: %s\n",
 						optarg);
 					exit(2);
 				}
+				/* convert sectors to K */
+				size /= 2;
+			}
+			continue;
+
+		case O(GROW,'Z'): /* array size */
+			if (array_size >= 0) {
+				fprintf(stderr, Name ": array-size may only be specified once. "
+					"Second value is %s.\n", optarg);
+				exit(2);
+			}
+			if (strcmp(optarg, "max") == 0)
+				array_size = 0;
+			else {
+				array_size = parse_size(optarg);
+				if (array_size <= 0) {
+					fprintf(stderr, Name ": invalid array size: %s\n",
+						optarg);
+					exit(2);
+				}
 			}
 			continue;
 
-		case O(GROW,'l'): /* hack - needed to understand layout */
+		case O(GROW,'l'):
 		case O(CREATE,'l'):
 		case O(BUILD,'l'): /* set raid level*/
 			if (level != UnSet) {
@@ -393,7 +437,10 @@ int main(int argc, char *argv[])
 					optarg);
 				exit(2);
 			}
-			if (level != 0 && level != -1 && level != 1 && level != -4 && level != -5 && mode == BUILD) {
+			if (level != 0 && level != LEVEL_LINEAR && level != 1 &&
+			    level != LEVEL_MULTIPATH && level != LEVEL_FAULTY &&
+			    level != 10 &&
+			    mode == BUILD) {
 				fprintf(stderr, Name ": Raid level %s not permitted with --build.\n",
 					optarg);
 				exit(2);
@@ -406,9 +453,18 @@ int main(int argc, char *argv[])
 			ident.level = level;
 			continue;
 
+		case O(GROW, 'p'): /* new layout */
+			if (layout_str) {
+				fprintf(stderr,Name ": layout may only be sent once.  "
+					"Second value was %s\n", optarg);
+				exit(2);
+			}
+			layout_str = optarg;
+			/* 'Grow' will parse the value */
+			continue;
+
 		case O(CREATE,'p'): /* raid5 layout */
 		case O(BUILD,'p'): /* faulty layout */
-		case O(GROW, 'p'): /* faulty reconfig */
 			if (layout != UnSet) {
 				fprintf(stderr,Name ": layout may only be sent once.  "
 					"Second value was %s\n", optarg);
@@ -424,7 +480,6 @@ int main(int argc, char *argv[])
 				exit(2);
 
 			case 5:
-			case 6:
 				layout = map_name(r5layout, optarg);
 				if (layout==UnSet) {
 					fprintf(stderr, Name ": layout %s not understood for raid5.\n",
@@ -432,40 +487,33 @@ int main(int argc, char *argv[])
 					exit(2);
 				}
 				break;
+			case 6:
+				layout = map_name(r6layout, optarg);
+				if (layout==UnSet) {
+					fprintf(stderr, Name ": layout %s not understood for raid6.\n",
+						optarg);
+					exit(2);
+				}
+				break;
 
 			case 10:
-				/* 'f', 'o' or 'n' followed by a number <= raid_disks */
-				if ((optarg[0] !=  'n' && optarg[0] != 'f' && optarg[0] != 'o') ||
-				    (copies = strtoul(optarg+1, &cp, 10)) < 1 ||
-				    copies > 200 ||
-				    *cp) {
+				layout = parse_layout_10(optarg);
+				if (layout < 0) {
 					fprintf(stderr, Name ": layout for raid10 must be 'nNN', 'oNN' or 'fNN' where NN is a number, not %s\n", optarg);
 					exit(2);
 				}
-				if (optarg[0] == 'n')
-					layout = 256 + copies;
-				else if (optarg[0] == 'o')
-					layout = 0x10000 + (copies<<8) + 1;
-				else
-					layout = 1 + (copies<<8);
 				break;
-			case -5: /* Faulty
-				  * modeNNN
-				  */
-
-			{
-				int ln = strcspn(optarg, "0123456789");
-				char *m = strdup(optarg);
-				int mode;
-				m[ln] = 0;
-				mode = map_name(faultylayout, m);
-				if (mode == UnSet) {
+			case LEVEL_FAULTY:
+				/* Faulty
+				 * modeNNN
+				 */
+				layout = parse_layout_faulty(optarg);
+				if (layout == -1) {
 					fprintf(stderr, Name ": layout %s not understood for faulty.\n",
 						optarg);
 					exit(2);
 				}
-				layout = mode | (atoi(optarg+ln)<< ModeShift);
-			}
+				break;
 			}
 			continue;
 
@@ -549,11 +597,16 @@ int main(int argc, char *argv[])
 
 		case O(CREATE,'N'):
 		case O(ASSEMBLE,'N'):
+		case O(MISC,'N'):
 			if (ident.name[0]) {
 				fprintf(stderr, Name ": name cannot be set twice.   "
 					"Second value %s.\n", optarg);
 				exit(2);
 			}
+			if (mode == MISC && !subarray) {
+				fprintf(stderr, Name ": -N/--name only valid with --update-subarray in misc mode\n");
+				exit(2);
+			}
 			if (strlen(optarg) > 32) {
 				fprintf(stderr, Name ": name '%s' is too long, 32 chars max.\n",
 					optarg);
@@ -580,11 +633,16 @@ int main(int argc, char *argv[])
 			continue;
 
 		case O(ASSEMBLE,'U'): /* update the superblock */
+		case O(MISC,'U'):
 			if (update) {
 				fprintf(stderr, Name ": Can only update one aspect of superblock, both %s and %s given.\n",
 					update, optarg);
 				exit(2);
 			}
+			if (mode == MISC && !subarray) {
+				fprintf(stderr, Name ": Only subarrays can be updated in misc mode\n");
+				exit(2);
+			}
 			update = optarg;
 			if (strcmp(update, "sparc2.2")==0)
 				continue;
@@ -602,6 +660,8 @@ int main(int argc, char *argv[])
 				continue;
 			if (strcmp(update, "devicesize")==0)
 				continue;
+			if (strcmp(update, "no-bitmap")==0)
+				continue;
 			if (strcmp(update, "byteorder")==0) {
 				if (ss) {
 					fprintf(stderr, Name ": must not set metadata type with --update=byteorder.\n");
@@ -628,15 +688,19 @@ int main(int argc, char *argv[])
 			}
 			fprintf(outf, "Valid --update options are:\n"
 		"     'sparc2.2', 'super-minor', 'uuid', 'name', 'resync',\n"
-		"     'summaries', 'homehost', 'byteorder', 'devicesize'.\n");
+		"     'summaries', 'homehost', 'byteorder', 'devicesize',\n"
+		"     'no-bitmap'\n");
 			exit(outf == stdout ? 0 : 2);
 
+		case O(INCREMENTAL,NoDegraded):
+			fprintf(stderr, Name ": --no-degraded is deprecated in Incremental mode\n");
 		case O(ASSEMBLE,NoDegraded): /* --no-degraded */
 			runstop = -1; /* --stop isn't allowed for --assemble,
 				       * so we overload slightly */
 			continue;
 
 		case O(ASSEMBLE,'c'): /* config file */
+		case O(INCREMENTAL, 'c'):
 		case O(MISC, 'c'):
 		case O(MONITOR,'c'):
 			if (configfile) {
@@ -671,6 +735,14 @@ int main(int argc, char *argv[])
 				program = optarg;
 			continue;
 
+		case O(MONITOR,'r'): /* rebuild increments */
+			increments = atoi(optarg);
+			if (increments>99 || increments<1) {
+				fprintf(stderr, Name ": please specify positive integer between 1 and 99 as rebuild increments.\n");
+				exit(2);
+			}
+			continue;
+
 		case O(MONITOR,'d'): /* delay in seconds */
 		case O(GROW, 'd'):
 		case O(BUILD,'d'): /* delay for bitmap updates */
@@ -724,6 +796,9 @@ int main(int argc, char *argv[])
 			devmode = 'r';
 			continue;
 		case O(MANAGE,'f'): /* set faulty */
+		case O(INCREMENTAL,'f'): /* r for incremental is taken, use f
+					  * even though we will both fail and
+					  * remove the device */
 			devmode = 'f';
 			continue;
 		case O(INCREMENTAL,'R'):
@@ -744,20 +819,8 @@ int main(int argc, char *argv[])
 			}
 			runstop = -1;
 			continue;
-
-		case O(MANAGE,'o'):
-			if (readonly < 0) {
-				fprintf(stderr, Name ": Cannot have both readonly and readwrite\n");
-				exit(2);
-			}
-			readonly = 1;
-			continue;
-		case O(MANAGE,'w'):
-			if (readonly > 0) {
-				fprintf(stderr, Name ": Cannot have both readwrite and readonly.\n");
-				exit(2);
-			}
-			readonly = -1;
+		case O(MANAGE,'t'):
+			test = 1;
 			continue;
 
 		case O(MISC,'Q'):
@@ -770,10 +833,23 @@ int main(int argc, char *argv[])
 		case O(MISC,'o'):
 		case O(MISC,'w'):
 		case O(MISC,'W'):
+		case O(MISC, Waitclean):
+		case O(MISC, DetailPlatform):
+		case O(MISC, KillSubarray):
+		case O(MISC, UpdateSubarray):
 			if (devmode && devmode != opt &&
 			    (devmode == 'E' || (opt == 'E' && devmode != 'Q'))) {
-				fprintf(stderr, Name ": --examine/-E cannot be given with -%c\n",
-					devmode =='E'?opt:devmode);
+				fprintf(stderr, Name ": --examine/-E cannot be given with ");
+				if (devmode == 'E') {
+					if (option_index >= 0)
+						fprintf(stderr, "--%s\n",
+							long_options[option_index].name);
+					else
+						fprintf(stderr, "-%c\n", opt);
+				} else if (isalpha(devmode))
+					fprintf(stderr, "-%c\n", devmode);
+				else
+					fprintf(stderr, "previous option\n");
 				exit(2);
 			}
 			devmode = opt;
@@ -835,7 +911,8 @@ int main(int argc, char *argv[])
 				continue;
 			}
 			/* probable typo */
-			fprintf(stderr, Name ": bitmap file must contain a '/', or be 'internal', or 'none'\n");
+			fprintf(stderr, Name ": bitmap file must contain a '/', or be 'internal', or 'none'\n"
+				"       not '%s'\n", optarg);
 			exit(2);
 
 		case O(GROW,BitmapChunk):
@@ -946,24 +1023,39 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");
 			exit(2);
 		}
-		if (mode == MANAGE || mode == GROW)
-			autof=1; /* Don't create */
-		mdfd = open_mddev(devlist->devname, autof);
-		if (mdfd < 0)
+		if (mode == MANAGE || mode == GROW) {
+			mdfd = open_mddev(devlist->devname, 1);
+			if (mdfd < 0)
+				exit(1);
+		} else
+			/* non-existent device is OK */
+			mdfd = open_mddev(devlist->devname, 0);
+		if (mdfd == -2) {
+			fprintf(stderr, Name ": device %s exists but is not an "
+				"md array.\n", devlist->devname);
 			exit(1);
+		}
 		if ((int)ident.super_minor == -2) {
 			struct stat stb;
+			if (mdfd < 0) {
+				fprintf(stderr, Name ": --super-minor=dev given, and "
+					"listed device %s doesn't exist.\n",
+					devlist->devname);
+				exit(1);
+			}
 			fstat(mdfd, &stb);
 			ident.super_minor = minor(stb.st_rdev);
 		}
+		if (mdfd >= 0 && mode != MANAGE && mode != GROW) {
+			/* We don't really want this open yet, we just might
+			 * have wanted to check some things
+			 */
+			close(mdfd);
+			mdfd = -1;
+		}
 	}
 
 	if (raiddisks) {
-		if (raiddisks > max_disks) {
-			fprintf(stderr, Name ": invalid number of raid devices: %d\n",
-				raiddisks);
-			exit(2);
-		}
 		if (raiddisks == 1 &&  !force && level != -5) {
 			fprintf(stderr, Name ": '1' is an unusual number of drives for an array, so it is probably\n"
 				"     a mistake.  If you really mean it you will need to specify --force before\n"
@@ -971,23 +1063,24 @@ int main(int argc, char *argv[])
 			exit(2);
 		}
 	}
-	if (sparedisks) {
-		if ( sparedisks > max_disks - raiddisks) {
-			fprintf(stderr, Name ": invalid number of spare-devices: %d\n",
-				sparedisks);
-			exit(2);
-		}
-	}
 
 	if (homehost == NULL)
-		homehost = conf_get_homehost();
-	if (homehost && strcmp(homehost, "<system>")==0) {
+		homehost = conf_get_homehost(&require_homehost);
+	if (homehost == NULL || strcmp(homehost, "<system>")==0) {
 		if (gethostname(sys_hostname, sizeof(sys_hostname)) == 0) {
 			sys_hostname[sizeof(sys_hostname)-1] = 0;
 			homehost = sys_hostname;
 		}
 	}
 
+	if ((mode != MISC || devmode != 'E') &&
+	    geteuid() != 0) {
+		fprintf(stderr, Name ": must be super-user to perform this action\n");
+		exit(1);
+	}
+
+	ident.autof = autof;
+
 	rv = 0;
 	switch(mode) {
 	case MANAGE:
@@ -996,7 +1089,7 @@ int main(int argc, char *argv[])
 			rv = Manage_ro(devlist->devname, mdfd, readonly);
 		if (!rv && devs_found>1)
 			rv = Manage_subdevs(devlist->devname, mdfd,
-					    devlist->next, verbose-quiet);
+					    devlist->next, verbose-quiet, test);
 		if (!rv && readonly < 0)
 			rv = Manage_ro(devlist->devname, mdfd, readonly);
 		if (!rv && runstop)
@@ -1011,22 +1104,23 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": %s not identified in config file.\n",
 					devlist->devname);
 				rv |= 1;
-			} else {
-				mdfd = open_mddev(devlist->devname,
-						  array_ident->autof ? array_ident->autof : autof);
-				if (mdfd < 0)
-					rv |= 1;
-				else {
-					rv |= Assemble(ss, devlist->devname, mdfd, array_ident,
-						       NULL, backup_file,
-						       readonly, runstop, update, homehost, verbose-quiet, force);
+				if (mdfd >= 0)
 					close(mdfd);
-				}
+			} else {
+				if (array_ident->autof == 0)
+					array_ident->autof = autof;
+				rv |= Assemble(ss, devlist->devname, array_ident,
+					       NULL, backup_file,
+					       readonly, runstop, update,
+					       homehost, require_homehost,
+					       verbose-quiet, force);
 			}
 		} else if (!scan)
-			rv = Assemble(ss, devlist->devname, mdfd, &ident,
+			rv = Assemble(ss, devlist->devname, &ident,
 				      devlist->next, backup_file,
-				      readonly, runstop, update, homehost, verbose-quiet, force);
+				      readonly, runstop, update,
+				      homehost, require_homehost,
+				      verbose-quiet, force);
 		else if (devs_found>0) {
 			if (update && devs_found > 1) {
 				fprintf(stderr, Name ": can only update a single array at a time\n");
@@ -1044,21 +1138,19 @@ int main(int argc, char *argv[])
 					rv |= 1;
 					continue;
 				}
-				mdfd = open_mddev(dv->devname,
-						  array_ident->autof ?array_ident->autof : autof);
-				if (mdfd < 0) {
-					rv |= 1;
-					continue;
-				}
-				rv |= Assemble(ss, dv->devname, mdfd, array_ident,
+				if (array_ident->autof == 0)
+					array_ident->autof = autof;
+				rv |= Assemble(ss, dv->devname, array_ident,
 					       NULL, backup_file,
-					       readonly, runstop, update, homehost, verbose-quiet, force);
-				close(mdfd);
+					       readonly, runstop, update,
+					       homehost, require_homehost,
+					       verbose-quiet, force);
 			}
 		} else {
-			mddev_ident_t array_list =  conf_get_ident(NULL);
+			mddev_ident_t a, array_list =  conf_get_ident(NULL);
 			mddev_dev_t devlist = conf_get_devs();
 			int cnt = 0;
+			int failures, successes;
 			if (devlist == NULL) {
 				fprintf(stderr, Name ": No devices listed in conf file were found.\n");
 				exit(1);
@@ -1071,29 +1163,41 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": --backup_file not meaningful with a --scan assembly.\n");
 				exit(1);
 			}
-			for (; array_list; array_list = array_list->next) {
-				mdu_array_info_t array;
-				mdfd = open_mddev(array_list->devname,
-						  array_list->autof ? array_list->autof : autof);
-				if (mdfd < 0) {
-					rv |= 1;
-					continue;
-				}
-				if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
-					/* already assembled, skip */
-					cnt++;
-				else {
-					rv |= Assemble(ss, array_list->devname, mdfd,
-						       array_list,
-						       NULL, NULL,
-						       readonly, runstop, NULL, homehost, verbose-quiet, force);
-					if (rv == 0) cnt++;
-				}
-				close(mdfd);
+			for (a = array_list; a ; a = a->next) {
+				a->assembled = 0;
+				if (a->autof == 0)
+					a->autof = autof;
 			}
-			if (homehost) {
+			do {
+				failures = 0;
+				successes = 0;
+				rv = 0;
+				for (a = array_list; a ; a = a->next) {
+					int r;
+					if (a->assembled)
+						continue;
+					if (a->devname &&
+					    strcasecmp(a->devname, "<ignore>") == 0)
+						continue;
+				
+					r = Assemble(ss, a->devname,
+						     a,
+						     NULL, NULL,
+						     readonly, runstop, NULL,
+						     homehost, require_homehost,
+						     verbose-quiet, force);
+					if (r == 0) {
+						a->assembled = 1;
+						successes++;
+					} else
+						failures++;
+					rv |= r;
+					cnt++;
+				}
+			} while (failures && successes);
+			if (homehost && cnt == 0) {
 				/* Maybe we can auto-assemble something.
-				 * Repeatedly call Assemble in auto-assmble mode
+				 * Repeatedly call Assemble in auto-assemble mode
 				 * until it fails
 				 */
 				int rv2;
@@ -1103,10 +1207,12 @@ int main(int argc, char *argv[])
 					mddev_dev_t devlist = conf_get_devs();
 					acnt = 0;
 					do {
-						rv2 = Assemble(ss, NULL, -1,
+						rv2 = Assemble(ss, NULL,
 							       &ident,
 							       devlist, NULL,
-							       readonly, runstop, NULL, homehost, verbose-quiet, force);
+							       readonly, runstop, NULL,
+							       homehost, require_homehost,
+							       verbose-quiet, force);
 						if (rv2==0) {
 							cnt++;
 							acnt++;
@@ -1119,15 +1225,18 @@ int main(int argc, char *argv[])
 					} while (rv2!=2);
 					/* Incase there are stacked devices, we need to go around again */
 				} while (acnt);
+#if 0
 				if (cnt == 0 && auto_update_home && homehost) {
 					/* Nothing found, maybe we need to bootstrap homehost info */
 					do {
 						acnt = 0;
 						do {
-							rv2 = Assemble(ss, NULL, -1,
+							rv2 = Assemble(ss, NULL,
 								       &ident,
 								       NULL, NULL,
-								       readonly, runstop, "homehost", homehost, verbose-quiet, force);
+								       readonly, runstop, "homehost",
+								       homehost, require_homehost,
+								       verbose-quiet, force);
 							if (rv2==0) {
 								cnt++;
 								acnt++;
@@ -1136,6 +1245,7 @@ int main(int argc, char *argv[])
 						/* Incase there are stacked devices, we need to go around again */
 					} while (acnt);
 				}
+#endif
 				if (cnt == 0 && rv == 0) {
 					fprintf(stderr, Name ": No arrays found in config file or automatically\n");
 					rv = 1;
@@ -1155,7 +1265,7 @@ int main(int argc, char *argv[])
 			break;
 		}
 		if (raiddisks == 0) {
-			fprintf(stderr, Name ": no raid-disks specified.\n");
+			fprintf(stderr, Name ": no raid-devices specified.\n");
 			rv = 1;
 			break;
 		}
@@ -1167,9 +1277,10 @@ int main(int argc, char *argv[])
 				break;
 			}
 		}
-		rv = Build(devlist->devname, mdfd, chunk, level, layout,
+		rv = Build(devlist->devname, chunk, level, layout,
 			   raiddisks, devlist->next, assume_clean,
-			   bitmap_file, bitmap_chunk, write_behind, delay, verbose-quiet);
+			   bitmap_file, bitmap_chunk, write_behind,
+			   delay, verbose-quiet, autof, size);
 		break;
 	case CREATE:
 		if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
@@ -1179,16 +1290,16 @@ int main(int argc, char *argv[])
 			break;
 		}
 		if (raiddisks == 0) {
-			fprintf(stderr, Name ": no raid-disks specified.\n");
+			fprintf(stderr, Name ": no raid-devices specified.\n");
 			rv = 1;
 			break;
 		}
 
-		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
+		rv = Create(ss, devlist->devname, chunk, level, layout, size<0 ? 0 : size,
 			    raiddisks, sparedisks, ident.name, homehost,
 			    ident.uuid_set ? ident.uuid : NULL,
 			    devs_found-1, devlist->next, runstop, verbose-quiet, force, assume_clean,
-			    bitmap_file, bitmap_chunk, write_behind, delay);
+			    bitmap_file, bitmap_chunk, write_behind, delay, autof);
 		break;
 	case MISC:
 		if (devmode == 'E') {
@@ -1207,24 +1318,51 @@ int main(int argc, char *argv[])
 			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief,
 				     export, scan,
 				     SparcAdjust, ss, homehost);
+		} else if (devmode == DetailPlatform) {
+			rv = Detail_Platform(ss ? ss->ss : NULL, ss ? scan : 1, verbose);
 		} else {
 			if (devlist == NULL) {
-				if (devmode=='D' && scan) {
-					/* apply --detail to all devices in /proc/mdstat */
+				if ((devmode=='D' || devmode == Waitclean) && scan) {
+					/* apply --detail or --wait-clean to
+					 * all devices in /proc/mdstat
+					 */
 					struct mdstat_ent *ms = mdstat_read(0, 1);
 					struct mdstat_ent *e;
+					struct map_ent *map = NULL;
+					int members;
+					int v = verbose>1?0:verbose+1;
+
+					for (members = 0; members <= 1; members++) {
 					for (e=ms ; e ; e=e->next) {
-						char *name = get_md_name(e->devnum);
+						char *name;
+						struct map_ent *me;
+						int member = e->metadata_version &&
+							strncmp(e->metadata_version,
+								"external:/", 10) == 0;
+						if (members != member)
+							continue;
+						me = map_by_devnum(&map, e->devnum);
+						if (me && me->path
+						    && strcmp(me->path, "/unknown") != 0)
+							name = me->path;
+						else
+							name = get_md_name(e->devnum);
 
 						if (!name) {
 							fprintf(stderr, Name ": cannot find device file for %s\n",
 								e->dev);
 							continue;
 						}
-						rv |= Detail(name, verbose>1?0:verbose+1,
-							     export, test, homehost);
+						if (devmode == 'D')
+							rv |= Detail(name, v,
+								     export, test,
+								     homehost);
+						else
+							rv |= WaitClean(name, -1, v);
 						put_md_name(name);
 					}
+					}
+					free_mdstat(ms);
 				} else	if (devmode == 'S' && scan) {
 					/* apply --stop to all devices in /proc/mdstat */
 					/* Due to possible stacking of devices, repeat until
@@ -1257,6 +1395,7 @@ int main(int argc, char *argv[])
 
 							put_md_name(name);
 						}
+						free_mdstat(ms);
 					} while (!last && err);
 					if (err) rv |= 1;
 				} else {
@@ -1272,13 +1411,37 @@ int main(int argc, char *argv[])
 						     export, test, homehost);
 					continue;
 				case 'K': /* Zero superblock */
-					rv |= Kill(dv->devname, force, quiet); continue;
+					if (ss)
+						rv |= Kill(dv->devname, ss, force, quiet,0);
+					else {
+						int q = quiet;
+						do {
+							rv |= Kill(dv->devname, NULL, force, q, 0);
+							q = 1;
+						} while (rv == 0);
+						rv &= ~2;
+					}
+					continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
 				case 'X':
 					rv |= ExamineBitmap(dv->devname, brief, ss); continue;
 				case 'W':
 					rv |= Wait(dv->devname); continue;
+				case Waitclean:
+					rv |= WaitClean(dv->devname, -1, verbose-quiet); continue;
+				case KillSubarray:
+					rv |= Kill_subarray(dv->devname, subarray, quiet);
+					continue;
+				case UpdateSubarray:
+					if (update == NULL) {
+						fprintf(stderr,
+							Name ": -U/--update must be specified with --update-subarray\n");
+						rv |= 1;
+						continue;
+					}
+					rv |= Update_subarray(dv->devname, subarray, update, &ident, quiet);
+					continue;
 				}
 				mdfd = open_mddev(dv->devname, 1);
 				if (mdfd>=0) {
@@ -1309,17 +1472,55 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
+		if (delay == 0) {
+			if (get_linux_version() > 20616)
+				/* mdstat responds to poll */
+				delay = 1000;
+			else
+				delay = 60;
+		}
 		rv= Monitor(devlist, mailaddr, program,
 			    delay?delay:60, daemonise, scan, oneshot,
-			    dosyslog, test, pidfile);
+			    dosyslog, test, pidfile, increments);
 		break;
 
 	case GROW:
+		if (array_size >= 0) {
+			/* alway impose array size first, independent of
+			 * anything else
+			 * Do not allow level or raid_disks changes at the
+			 * same time as that can be irreversibly destructive.
+			 */
+			struct mdinfo sra;
+			int err;
+			if (raiddisks || level != UnSet) {
+				fprintf(stderr, Name ": cannot change array size in same operation "
+					"as changing raiddisks or level.\n"
+					"    Change size first, then check that data is still intact.\n");
+				rv = 1;
+				break;
+			}
+			sysfs_init(&sra, mdfd, 0);
+			if (array_size == 0)
+				err = sysfs_set_str(&sra, NULL, "array_size", "default");
+			else
+				err = sysfs_set_num(&sra, NULL, "array_size", array_size / 2);
+			if (err < 0) {
+				if (errno == E2BIG)
+					fprintf(stderr, Name ": --array-size setting"
+						" is too large.\n");
+				else
+					fprintf(stderr, Name ": current kernel does"
+						" not support setting --array-size\n");
+				rv = 1;
+				break;
+			}
+		}
 		if (devs_found > 1) {
 
 			/* must be '-a'. */
-			if (size >= 0 || raiddisks) {
-				fprintf(stderr, Name ": --size, --raiddisks, and --add are exclusing in --grow mode\n");
+			if (size >= 0 || raiddisks || chunk || layout_str != NULL || bitmap_file) {
+				fprintf(stderr, Name ": --add cannot be used with other geometry changes in --grow mode\n");
 				rv = 1;
 				break;
 			}
@@ -1328,20 +1529,21 @@ int main(int argc, char *argv[])
 				if (rv)
 					break;
 			}
-		} else if ((size >= 0) + (raiddisks != 0) +  (layout != UnSet) + (bitmap_file != NULL)> 1) {
-			fprintf(stderr, Name ": can change at most one of size, raiddisks, bitmap, and layout\n");
-			rv = 1;
-			break;
-		} else if (layout != UnSet)
-			rv = Manage_reconfig(devlist->devname, mdfd, layout);
-		else if (size >= 0 || raiddisks)
-			rv = Grow_reshape(devlist->devname, mdfd, quiet, backup_file,
-					  size, level, layout, chunk, raiddisks);
-		else if (bitmap_file) {
-			if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
+		} else if (bitmap_file) {
+			if (size >= 0 || raiddisks || chunk || layout_str != NULL) {
+				fprintf(stderr, Name ": --bitmap changes cannot be used with other geometry changes in --grow mode\n");
+				rv = 1;
+				break;
+			}
+			if (delay == 0)
+				delay = DEFAULT_BITMAP_DELAY;
 			rv = Grow_addbitmap(devlist->devname, mdfd, bitmap_file,
 					    bitmap_chunk, delay, write_behind, force);
-		} else
+		} else if (size >= 0 || raiddisks != 0 || layout_str != NULL
+			   || chunk != 0 || level != UnSet) {
+			rv = Grow_reshape(devlist->devname, mdfd, quiet, backup_file,
+					  size, level, layout_str, chunk, raiddisks);
+		} else if (array_size < 0)
 			fprintf(stderr, Name ": no changes to --grow\n");
 		break;
 	case INCREMENTAL:
@@ -1354,6 +1556,11 @@ int main(int argc, char *argv[])
 			 ": --incremental --scan meaningless without --run.\n");
 				break;
 			}
+			if (devmode == 'f') {
+				fprintf(stderr, Name
+			 ": --incremental --scan --fail not supported.\n");
+				break;
+			}
 			rv = IncrementalScan(verbose);
 		}
 		if (!devlist) {
@@ -1370,8 +1577,12 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
+		if (devmode == 'f') {
+			rv = IncrementalRemove(devlist->devname, verbose-quiet);
+			break;
+		}
 		rv = Incremental(devlist->devname, verbose-quiet, runstop,
-				 ss, homehost, autof);
+				 ss, homehost, require_homehost, autof);
 		break;
 	case AUTODETECT:
 		autodetect();
